/*----- PROTECTED REGION ID(E2Corrector.cpp) ENABLED START -----*/
/* clang-format on */
//=============================================================================
//
// file :        E2Corrector.cpp
//
// description : C++ source for the E2Corrector class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               E2Corrector are implemented in this file.
//
// project :     Elettra 2.0 pure corrector magnet device
//
// This file is part of Tango device class.
//
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
//
//
// Elettra - Sincrotrone Trieste S.c.p.A.
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include "E2Corrector.h"
#include "E2CorrectorClass.h"
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	E2Corrector.cpp

/**
 *  E2Corrector class description:
 *    Elettra 2.0 corrector magnet device
 *    Use calibration table handle current/strgenght converstions.
 *    The magnet device is connected to a power supply tango device , real or simulated.
 *    
 *    Based on ESRF EBSmagnet concepts and libraries.
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name            |  Method name
//================================================================
//  State                   |  Inherited (no method)
//  Status                  |  Inherited (no method)
//  Reset                   |  reset
//  ResetResonanceStrength  |  reset_resonance_strength
//  ApplyDesignStrength     |  apply_design_strength
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  Strength            |  Tango::DevDouble	Scalar
//  CorrectionStrength  |  Tango::DevDouble	Scalar
//  ResonanceStrength   |  Tango::DevDouble	Scalar
//  Current             |  Tango::DevDouble	Scalar
//  BeamEnergy          |  Tango::DevDouble	Scalar
//  DesignStrength      |  Tango::DevDouble	Scalar
//================================================================

namespace E2Corrector_ns
{
/*----- PROTECTED REGION ID(E2Corrector::namespace_starting) ENABLED START -----*/
/* clang-format on */
//	static initializations
/* clang-format off */
/*----- PROTECTED REGION END -----*/	//	E2Corrector::namespace_starting

//--------------------------------------------------------
/**
 *	Method     : E2Corrector::E2Corrector()
 *	Description: Constructors for a Tango device
 *                implementing the classE2Corrector
 */
//--------------------------------------------------------
E2Corrector::E2Corrector(Tango::DeviceClass *cl, std::string &s)
 : TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(E2Corrector::constructor_1) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::constructor_1
}
//--------------------------------------------------------
E2Corrector::E2Corrector(Tango::DeviceClass *cl, const char *s)
 : TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(E2Corrector::constructor_2) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::constructor_2
}
//--------------------------------------------------------
E2Corrector::E2Corrector(Tango::DeviceClass *cl, const char *s, const char *d)
 : TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(E2Corrector::constructor_3) ENABLED START -----*/
	/* clang-format on */
	init_device();
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::constructor_3
}
//--------------------------------------------------------
E2Corrector::~E2Corrector()
{
	delete_device();
}

//--------------------------------------------------------
/**
 *	Method     : E2Corrector::delete_device()
 *	Description: will be called at device destruction or at init command
 */
//--------------------------------------------------------
void E2Corrector::delete_device()
{
	DEBUG_STREAM << "E2Corrector::delete_device() " << device_name << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::delete_device) ENABLED START -----*/
	/* clang-format on */
	//	Delete device allocated objects

	if(state_event_id) current_dev->unsubscribe_event(state_event_id);
	if(current_set_event_id) current_dev->unsubscribe_event(current_set_event_id);
	if(current_event_id) current_dev->unsubscribe_event(current_event_id);
	if(current_cb) delete current_cb;
	delete attr_Current_read;
	if (current_dev) delete current_dev;
	if (corrector) delete corrector;
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::delete_device
	delete[] attr_Strength_read;
	delete[] attr_CorrectionStrength_read;
	delete[] attr_ResonanceStrength_read;
	delete[] attr_BeamEnergy_read;
	delete[] attr_DesignStrength_read;
}

//--------------------------------------------------------
/**
 *	Method     : E2Corrector::init_device()
 *	Description: will be called at device initialization.
 */
//--------------------------------------------------------
void E2Corrector::init_device()
{
	DEBUG_STREAM << "E2Corrector::init_device() create device " << device_name << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::init_device_before) ENABLED START -----*/
	/* clang-format on */
	//	Initialization before get_device_property() call
	/* clang-format off */

	set_state(Tango::INIT);
	set_status("Init");
	attr_Current_read = new Tango::DevDouble[1];
	design_strength_initialized = false;
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::init_device_before


	//	Get the device properties from database
	get_device_property();

	attr_Strength_read = new Tango::DevDouble[1];
	attr_CorrectionStrength_read = new Tango::DevDouble[1];
	attr_ResonanceStrength_read = new Tango::DevDouble[1];
	attr_BeamEnergy_read = new Tango::DevDouble[1];
	attr_DesignStrength_read = new Tango::DevDouble[1];
	//	No longer if mandatory property not set.
	if (mandatoryNotDefined)
		return;

	/*----- PROTECTED REGION ID(E2Corrector::init_device) ENABLED START -----*/
	/* clang-format on */
	//	Initialize device
	/* clang-format off */
	current_dev = nullptr;
	current_cb = nullptr;
	current_event_id = 0;
	current_set_event_id = 0;
	state_event_id = 0;
	set_current.resize(1, 0.0);
	set_strength.resize(1,0.0);
	try{
		current_dev = new Tango::DeviceProxy(currentDevice);
	}
	catch (Tango::DevFailed &ex){
		ERROR_STREAM << "failed to create Tango device proxy for "
				<< currentDevice;
		set_state(Tango::UNKNOWN);
		std::string st = get_status();
		st += "\nfailed to create Tango device proxy for CurrentDevice";
		set_status(st);
		return;
	}

	//create MagnetModel::Corrector - handles calibration tables and interpolations
	// we do not use teh EBS shehme with global properties... yet
	//TODO: fix/determine format and handling of Elettra 2.0 calibrations

	corrector = nullptr;
	try {
		std::string meas_strength_file_name = calibrationPath + "/" + calibrationFileName;
		std::string param_file_name = calibrationPath + "/" + parameterFileName;
		corrector = new MagnetModel::Corrector();
		corrector->init(true,1.0, meas_strength_file_name, param_file_name, magnetName);
	} catch (std::exception &e) {
		std::cerr << "fatalerror "<< e.what() << std::endl;
		ERROR_STREAM << "failed to create corrector model " << currentDevice;
		set_state(Tango::UNKNOWN);
		set_status("failed to create corrector model CurrentDevice");
		return;
	}

	// init energy
	*attr_BeamEnergy_read = default_beam_energy;
	Tango::WAttribute &cat = get_device_attr()->get_w_attr_by_name("BeamEnergy");
	cat.set_write_value(default_beam_energy);
	calc_rigidity(default_beam_energy);

	//get WAtrirbute for handling cross dependencies
	strengthAtt =  &get_device_attr()->get_w_attr_by_name("Strength");
	correctionAtt = &get_device_attr()->get_w_attr_by_name("CorrectionStrength");
	resonanceAtt = &get_device_attr()->get_w_attr_by_name("ResonanceStrength");
	cursetAtt = &(get_device_attr()->get_w_attr_by_name("Current"));
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::init_device
}

//--------------------------------------------------------
/**
 *	Method     : E2Corrector::get_device_property()
 *	Description: Read database to initialize property data members.
 */
//--------------------------------------------------------
void E2Corrector::get_device_property()
{
	/*----- PROTECTED REGION ID(E2Corrector::get_device_property_before) ENABLED START -----*/
	/* clang-format on */
	//	Initialize property data members
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::get_device_property_before

	mandatoryNotDefined = false;

	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("CurrentDevice"));
	dev_prop.push_back(Tango::DbDatum("Model"));
	dev_prop.push_back(Tango::DbDatum("SerialNumber"));
	dev_prop.push_back(Tango::DbDatum("MagnetName"));
	dev_prop.push_back(Tango::DbDatum("CalibrationFileName"));
	dev_prop.push_back(Tango::DbDatum("CalibrationPath"));
	dev_prop.push_back(Tango::DbDatum("ParameterFileName"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);

		//	get instance on E2CorrectorClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		E2CorrectorClass	*ds_class =
			(static_cast<E2CorrectorClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize CurrentDevice from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  currentDevice;
		else {
			//	Try to initialize CurrentDevice from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  currentDevice;
		}
		//	And try to extract CurrentDevice value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  currentDevice;

		//	Try to initialize Model from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  model;
		else {
			//	Try to initialize Model from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  model;
		}
		//	And try to extract Model value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  model;

		//	Try to initialize SerialNumber from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  serialNumber;
		else {
			//	Try to initialize SerialNumber from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  serialNumber;
		}
		//	And try to extract SerialNumber value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  serialNumber;

		//	Try to initialize MagnetName from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  magnetName;
		else {
			//	Try to initialize MagnetName from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  magnetName;
		}
		//	And try to extract MagnetName value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  magnetName;

		//	Try to initialize CalibrationFileName from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  calibrationFileName;
		else {
			//	Try to initialize CalibrationFileName from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  calibrationFileName;
		}
		//	And try to extract CalibrationFileName value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  calibrationFileName;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize CalibrationPath from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  calibrationPath;
		else {
			//	Try to initialize CalibrationPath from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  calibrationPath;
		}
		//	And try to extract CalibrationPath value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  calibrationPath;

		//	Try to initialize ParameterFileName from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  parameterFileName;
		else {
			//	Try to initialize ParameterFileName from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  parameterFileName;
		}
		//	And try to extract ParameterFileName value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  parameterFileName;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

	}

	/*----- PROTECTED REGION ID(E2Corrector::get_device_property_after) ENABLED START -----*/
	/* clang-format on */
	//	Check device property data members init
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::get_device_property_after
}
//--------------------------------------------------------
/**
 *	Method     : E2Corrector::check_mandatory_property()
 *	Description: For mandatory properties check if defined in database.
 */
//--------------------------------------------------------
void E2Corrector::check_mandatory_property(Tango::DbDatum &class_prop, Tango::DbDatum &dev_prop)
{
	//	Check if all properties are empty
	if (class_prop.is_empty() && dev_prop.is_empty())
	{
		TangoSys_OMemStream	tms;
		tms << std::endl <<"Property \'" << dev_prop.name;
		if (Tango::Util::instance()->_UseDb==true)
			tms << "\' is mandatory but not defined in database";
		else
			tms << "\' is mandatory but cannot be defined without database";
		append_status(tms.str());
		mandatoryNotDefined = true;
		/*----- PROTECTED REGION ID(E2Corrector::check_mandatory_property) ENABLED START -----*/
		/* clang-format on */
		std::cerr << tms.str() << " for " << device_name << std::endl;
		/* clang-format off */
		/*----- PROTECTED REGION END -----*/	//	E2Corrector::check_mandatory_property
	}
}


//--------------------------------------------------------
/**
 *	Method     : E2Corrector::always_executed_hook()
 *	Description: method always executed before any command is executed
 */
//--------------------------------------------------------
void E2Corrector::always_executed_hook()
{
	DEBUG_STREAM << "E2Corrector::always_executed_hook()  " << device_name << std::endl;
	if (mandatoryNotDefined)
	{
		Tango::Except::throw_exception(
					(const char *)"PROPERTY_NOT_SET",
					get_status().c_str(),
					(const char *)"E2Corrector::always_executed_hook()");
	}
	/*----- PROTECTED REGION ID(E2Corrector::always_executed_hook) ENABLED START -----*/
	/* clang-format on */
	//	code always executed before all requests
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method     : E2Corrector::read_attr_hardware()
 *	Description: Hardware acquisition for attributes
 */
//--------------------------------------------------------
void E2Corrector::read_attr_hardware(TANGO_UNUSED(std::vector<long> &attr_list))
{
	DEBUG_STREAM << "E2Corrector::read_attr_hardware(std::vector<long> &attr_list) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::read_attr_hardware) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::read_attr_hardware
}
//--------------------------------------------------------
/**
 *	Method     : E2Corrector::write_attr_hardware()
 *	Description: Hardware writing for attributes
 */
//--------------------------------------------------------
void E2Corrector::write_attr_hardware(TANGO_UNUSED(std::vector<long> &attr_list))
{
	DEBUG_STREAM << "E2Corrector::write_attr_hardware(std::vector<long> &attr_list) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::write_attr_hardware) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::write_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute Strength related method
 *	Description: corrector strength 1/m^3 ( m-3)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E2Corrector::read_Strength(Tango::Attribute &attr)
{
	DEBUG_STREAM << "E2Corrector::read_Strength(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::read_Strength) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_Strength_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::read_Strength
}
//--------------------------------------------------------
/**
 *	Write attribute Strength related method
 *	Description: corrector strength 1/m^3 ( m-3)
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E2Corrector::write_Strength(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "E2Corrector::write_Strength(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(E2Corrector::write_Strength) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	omni_mutex_lock guardio(iomutex);
	std::vector<double> req_strength;
	std::vector<double> curr_out;
	req_strength.push_back(w_val);
	corrector->compute_currents(magnet_rigidity, req_strength, curr_out);
	Tango::DeviceAttribute watt("Current",curr_out[0]);
	current_dev->write_attribute(watt);
	set_strength[0] = w_val;
	wValCorrStrength = w_val - wValResStrength;
	correctionAtt->set_write_value(wValCorrStrength);
	//catch(std::excpetion &ex){}
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::write_Strength
}
//--------------------------------------------------------
/**
 *	Read attribute CorrectionStrength related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E2Corrector::read_CorrectionStrength(Tango::Attribute &attr)
{
	DEBUG_STREAM << "E2Corrector::read_CorrectionStrength(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::read_CorrectionStrength) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_CorrectionStrength_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::read_CorrectionStrength
}
//--------------------------------------------------------
/**
 *	Write attribute CorrectionStrength related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E2Corrector::write_CorrectionStrength(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "E2Corrector::write_CorrectionStrength(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(E2Corrector::write_CorrectionStrength) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/* TODO : implement is_set_point valid and exploit it!
			if(!is_setpoint_valid(w_val)) {
			    attr.rollback();
			    return;
			  }
	 */
	omni_mutex_lock guardio(iomutex);
	Tango::DevDouble newStrength;
	//resonanceAtt->get_write_value(_resStrength);
	newStrength = w_val + wValResStrength + *attr_DesignStrength_read;
	std::vector<double> req_strength;
	std::vector<double> curr_out;
	req_strength.push_back(newStrength);
	corrector->compute_currents(magnet_rigidity, req_strength, curr_out);
	Tango::DeviceAttribute watt("Current", curr_out[0]);
	current_dev->write_attribute(watt);
	wValCorrStrength = w_val;
	set_strength[0] = newStrength;
	strengthAtt->set_write_value(set_strength[0]);

	/*----- PROTECTED REGION END -----*/	//	E2Corrector::write_CorrectionStrength
}
//--------------------------------------------------------
/**
 *	Read attribute ResonanceStrength related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E2Corrector::read_ResonanceStrength(Tango::Attribute &attr)
{
	DEBUG_STREAM << "E2Corrector::read_ResonanceStrength(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::read_ResonanceStrength) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_ResonanceStrength_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::read_ResonanceStrength
}
//--------------------------------------------------------
/**
 *	Write attribute ResonanceStrength related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E2Corrector::write_ResonanceStrength(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "E2Corrector::write_ResonanceStrength(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(E2Corrector::write_ResonanceStrength) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	/* TODO : implement is_set_point valid and exploit it!
				if(!is_setpoint_valid(w_val)) {
				    attr.rollback();
				    return;
				  }
	 */
	omni_mutex_lock guardio(iomutex);
	Tango::DevDouble newStrength;
	newStrength = w_val + wValCorrStrength + *attr_DesignStrength_read;
	std::vector<double> req_strength;
	std::vector<double> curr_out;
	req_strength.push_back(newStrength);
	corrector->compute_currents(magnet_rigidity, req_strength, curr_out);
	Tango::DeviceAttribute watt("Current", curr_out[0]);
	current_dev->write_attribute(watt);
	wValResStrength = w_val;
	*attr_ResonanceStrength_read = w_val;
	set_strength[0] = newStrength;
	strengthAtt->set_write_value(set_strength[0]);
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::write_ResonanceStrength
}
//--------------------------------------------------------
/**
 *	Read attribute Current related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E2Corrector::read_Current(Tango::Attribute &attr)
{
	DEBUG_STREAM << "E2Corrector::read_Current(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::read_Current) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_Current_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::read_Current
}
//--------------------------------------------------------
/**
 *	Write attribute Current related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E2Corrector::write_Current(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "E2Corrector::write_Current(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(E2Corrector::write_Current) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	Tango::DeviceAttribute watt("Current",w_val);
	current_dev->write_attribute(watt);
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::write_Current
}
//--------------------------------------------------------
/**
 *	Read attribute BeamEnergy related method
 *	Description: Electron beam energy in eV. Used to calculate magnetic rigidity in strength calculations.
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E2Corrector::read_BeamEnergy(Tango::Attribute &attr)
{
	DEBUG_STREAM << "E2Corrector::read_BeamEnergy(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::read_BeamEnergy) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_BeamEnergy_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::read_BeamEnergy
}
//--------------------------------------------------------
/**
 *	Write attribute BeamEnergy related method
 *	Description: Electron beam energy in eV. Used to calculate magnetic rigidity in strength calculations.
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E2Corrector::write_BeamEnergy(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "E2Corrector::write_BeamEnergy(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(E2Corrector::write_BeamEnergy) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	*attr_BeamEnergy_read = w_val;
	calc_rigidity(w_val);
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::write_BeamEnergy
}
//--------------------------------------------------------
/**
 *	Read attribute DesignStrength related method
 *	Description: corrector design strength 1/m^3 ( m-3). The strenght reqiured by the nominal optic
 *               The attribute is memorized , expert only.
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E2Corrector::read_DesignStrength(Tango::Attribute &attr)
{
	DEBUG_STREAM << "E2Corrector::read_DesignStrength(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::read_DesignStrength) ENABLED START -----*/
	/* clang-format on */
	//	Set the attribute value
	attr.set_value(attr_DesignStrength_read);
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::read_DesignStrength
}
//--------------------------------------------------------
/**
 *	Write attribute DesignStrength related method
 *	Description: corrector design strength 1/m^3 ( m-3). The strenght reqiured by the nominal optic
 *               The attribute is memorized , expert only.
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E2Corrector::write_DesignStrength(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "E2Corrector::write_DesignStrength(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(E2Corrector::write_DesignStrength) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code
	/* clang-format off */
	*attr_DesignStrength_read = w_val;
	if (design_strength_initialized == false){ // only with a value of DesignStrength init process can be completed
		design_strength_initialized = true;
		init_strengths();
		subscribe_events();
	}
	else init_strengths();
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::write_DesignStrength
}

//--------------------------------------------------------
/**
 *	Method     : E2Corrector::add_dynamic_attributes()
 *	Description: Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void E2Corrector::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(E2Corrector::add_dynamic_attributes) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code to create and add dynamic attributes if any
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command Reset related method
 *	Description: reset the corrector
 *
 */
//--------------------------------------------------------
void E2Corrector::reset()
{
	DEBUG_STREAM << "E2Corrector::Reset()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::reset) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::reset
}
//--------------------------------------------------------
/**
 *	Command ResetResonanceStrength related method
 *	Description: Transfer ResonanceStranght to CorrectionStrength and set ResonanceStrength to 0. Strngth is not modified
 *
 */
//--------------------------------------------------------
void E2Corrector::reset_resonance_strength()
{
	DEBUG_STREAM << "E2Corrector::ResetResonanceStrength()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::reset_resonance_strength) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code

	/* clang-format off */
	//transfer resonance strength to correction strength and than force to 0 resonance strength
	omni_mutex_lock guardio(iomutex);

	wValCorrStrength +=  wValResStrength;
	wValResStrength = 0.0;

	correctionAtt->set_write_value(wValCorrStrength);
	*attr_CorrectionStrength_read = wValCorrStrength;
	resonanceAtt->set_write_value(0.0);
	*attr_ResonanceStrength_read = 0.0;
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::reset_resonance_strength
}
//--------------------------------------------------------
/**
 *	Command ApplyDesignStrength related method
 *	Description: Apply the memorized DesignStrength strenght to the magnet.
 *
 */
//--------------------------------------------------------
void E2Corrector::apply_design_strength()
{
	DEBUG_STREAM << "E2Corrector::ApplyDesignStrength()  - " << device_name << std::endl;
	/*----- PROTECTED REGION ID(E2Corrector::apply_design_strength) ENABLED START -----*/
	/* clang-format on */

	//	Add your own code

	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::apply_design_strength
}
//--------------------------------------------------------
/**
 *	Method     : E2Corrector::add_dynamic_commands()
 *	Description: Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void E2Corrector::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(E2Corrector::add_dynamic_commands) ENABLED START -----*/
	/* clang-format on */
	//	Add your own code to create and add dynamic commands if any
	/* clang-format off */
	/*----- PROTECTED REGION END -----*/	//	E2Corrector::add_dynamic_commands
}

/*----- PROTECTED REGION ID(E2Corrector::namespace_ending) ENABLED START -----*/
/* clang-format on */
//	Additional Methods
/* clang-format off */
//--------------------------------------------------------
void E2Corrector::calc_rigidity(const double& energy)
{
	//claculate magnet ridity for electrons at energy (a.k.a. Brho)
	magnet_rigidity = sqrt(energy*energy - E0*E0)/c_light;
	magnet_rigidity_inv = 1.0/magnet_rigidity;
	//std::cout << magnet_rigidity << " simplified:" << (energy/1e9)*3.3356 << std::endl;
}

void E2Corrector::set_current_write(const double& curr)
{
	cursetAtt->set_write_value(curr);
	set_current[0] = curr;
}
// ------------------------------------------------------------
void E2Corrector::init_strengths(void) // initialize strengths from power supply value TODO: separate read and set values
{
	Tango::DeviceAttribute da;
	try {
		da = current_dev->read_attribute("Current");
		vector<Tango::DevDouble> outval;
		da >> outval;
		std::vector<Tango::DevDouble> _current;
		std::vector<Tango::DevDouble> _strength;
		_current.push_back(outval[1]); //use set value of current
		corrector->compute_strengths(magnet_rigidity_inv, _current, _strength);

		cursetAtt->set_write_value(_current[0]);
		set_current[0] = _current[0];

		strengthAtt->set_write_value(_strength[0]);
		*attr_Strength_read = _strength[0];
		set_strength[0] = _strength[0];
        Tango::DevDouble corr = _strength[0] - *attr_DesignStrength_read;

		correctionAtt->set_write_value(corr);
		*attr_CorrectionStrength_read = corr;
		wValCorrStrength = corr;

		resonanceAtt->set_write_value(0.0); // start with resonance strength always at 0!
		*attr_ResonanceStrength_read = 0.0;
		wValResStrength = 0.0;

	} catch (Tango::DevFailed &e) {
		Tango::Except::print_exception(e);
	}
}
//------------------------------------------------------------
// handle changes poer poer supply read current (actula current)
void E2Corrector::update_values_read(void)
{
	if (isnan(*attr_Current_read)){ //handle error
		attr_Strength_read[0] = NAN;
		return;
	}
	//INFO_STREAM << "update values read";
	std::vector<Tango::DevDouble> read_curr;
	std::vector<Tango::DevDouble> read_strength;
	omni_mutex_lock guardio(iomutex);
	read_curr.push_back(*attr_Current_read);
	corrector->compute_strengths(magnet_rigidity_inv, read_curr, read_strength);
	attr_Strength_read[0] =  read_strength[0];
	attr_CorrectionStrength_read[0] = attr_Strength_read[0] -( attr_DesignStrength_read[0] +  attr_ResonanceStrength_read[0]);
}

//------------------------------------------------------------
// hndle change of write set point of power supply cunnrent
void E2Corrector::update_values_write(void)
{
	if (isnan(*attr_Current_read)){ //handle error
		attr_Strength_read[0] = NAN;
		return;
	}

	INFO_STREAM << "update values write";
	std::vector<Tango::DevDouble> read_curr;
	std::vector<Tango::DevDouble> read_strength;
	omni_mutex_lock guardio(iomutex);
	read_curr.push_back(set_current[0]); //must use the write value
	corrector->compute_strengths(magnet_rigidity_inv, read_curr, read_strength);
	wValCorrStrength = read_strength[0] - (wValResStrength + *attr_DesignStrength_read);
	correctionAtt->set_write_value(wValCorrStrength);
}

//------------------------------------------------------------
// handel che of power supply state - may just filter values
void E2Corrector::update_values_state(void)
{
	//handle state - status
	//INFO_STREAM << "update values state";
	switch (get_state()) {
	case Tango::ON:
		set_status("On");
		break;
	case Tango::OFF:
		set_status("Off");
		break;
	case Tango::FAULT:
		set_status("Fault");
		break;
	case Tango::RUNNING:
		set_status("Running");
		break;
	case Tango::MOVING:
		set_status("Moving");
		break;
	default:
		set_status("Unknown");
	}

}

void E2Corrector::subscribe_events(void)
{
	try{
			current_cb = new CurrentEventCallback(this);
			current_event_id = current_dev->subscribe_event("current",Tango::CHANGE_EVENT,current_cb);
			current_set_event_id = current_dev->subscribe_event("currentSet",Tango::CHANGE_EVENT,current_cb);
			state_event_id = current_dev->subscribe_event("State",Tango::CHANGE_EVENT,current_cb);
		}
		catch(Tango::DevFailed &ex){
			current_cb = nullptr;
			ERROR_STREAM << "failed to subscribe Tango  events for " << currentDevice;
			set_state(Tango::UNKNOWN);
			set_status("failed to subscribe Tango  events for CurrentDevice");
			return;
		}
}
/* trick to start sunbscrition afetr some delay
void E2Corrector::delayed_subscribe(int delayMs)
#include <thread>
#include <chrono>
{
	std::thread([this, delayMs]() {
		std::this_thread::sleep_for(std::chrono::milliseconds(delayMs));
		this->subscribe_events();  // Call the method after the delay
	}).detach();  // Detach the thread to run independently
}
*/
//------------------------------------------------------------
//------------------------------------------------------------
CurrentEventCallback::CurrentEventCallback(E2Corrector* quad) //callback constructor
{
	e2_quad = quad;
}
void CurrentEventCallback::push_event(Tango::EventData* ev) //callback core method
{
	bool update_read = false;
	bool update_write = false;
	bool update_state = false;
	if (ev->err) {
		//std::cerr << ev->err << " " << ev->attr_name << std::endl;
		e2_quad->set_state(Tango::FAULT);
		e2_quad->set_status("event error");
		*e2_quad->attr_Current_read = NAN;
		e2_quad->update_values_read();
		return;
	}
	{
		omni_mutex_lock guardio(e2_quad->iomutex);
		//TODO add method to update the corrector when on value changes
		std::string fullattrname = ev->attr_name;
		std::size_t pos = fullattrname.find_last_of("/");
		std::string attrname = fullattrname.substr(pos + 1);
		if (attrname == "current") {
			*(ev->attr_value) >> *(e2_quad->attr_Current_read);
			update_read = true;
		}
		if (attrname == "currentset") {
			Tango::DevDouble newcurr;
			*(ev->attr_value) >> newcurr;
			e2_quad->set_current_write(newcurr);
			update_write = true;
		}
		if (attrname == "state") {
			Tango::DevState st;
			*(ev->attr_value) >> st;
			e2_quad->set_state(st);
			update_state = true;
		}
	}
	if (update_read) e2_quad->update_values_read();
	if (update_write) e2_quad->update_values_write();
	if (update_state) e2_quad->update_values_state();
}

/*----- PROTECTED REGION END -----*/	//	E2Corrector::namespace_ending
} //	namespace
